/* Copyright (c) 2006-2016. The SimGrid Team. All rights reserved.          */

/* This program is free software; you can redistribute it and/or modify it
 * under the terms of the license (GNU LGPL) which comes with this package. */

#ifndef SIMGRID_S4U_COND_VARIABLE_HPP
#define SIMGRID_S4U_COND_VARIABLE_HPP

#include <condition_variable>
#include <mutex>
#include <utility> // std::swap

#include <simgrid/simix.h>
#include <simgrid/s4u/mutex.hpp>

namespace simgrid {
namespace s4u {

class Mutex;

/** A condition variable
 *
 *  This is based on std::condition_variable and should respect the same
 *  semantic. But we currently use (only) double for both durations and
 *  timestamp timeouts.
 */
XBT_PUBLIC_CLASS ConditionVariable {
  
public:
  ConditionVariable();

  ConditionVariable(ConditionVariable* cond) : cond_(SIMIX_cond_ref(cond->cond_)) {}
  ~ConditionVariable();

  // Copy+move (with the copy-and-swap idiom):
  ConditionVariable(ConditionVariable const& cond) : cond_(SIMIX_cond_ref(cond.cond_)) {}
  friend void swap(ConditionVariable& first, ConditionVariable& second)
  {
    using std::swap;
    swap(first.cond_, second.cond_);
  }
  ConditionVariable& operator=(ConditionVariable cond)
  {
    swap(*this, cond);
    return *this;
  }
  ConditionVariable(ConditionVariable&& cond) : cond_(nullptr)
  {
    swap(*this, cond);
  }

  bool valid() const
  {
    return cond_ != nullptr;
  }
  
  //  Wait functions:

  void wait(std::unique_lock<Mutex>& lock);
  std::cv_status wait_until(std::unique_lock<Mutex>& lock, double timeout_time);
  std::cv_status wait_for(std::unique_lock<Mutex>& lock, double duration);

  // Variants which takes a predicate:

  template<class P>
  void wait(std::unique_lock<Mutex>& lock, P pred)
  {
    while (!pred())
      wait(lock);
  }
  template<class P>
  bool wait_until(std::unique_lock<Mutex>& lock, double timeout_time, P pred)
  {
    while (!pred())
      if (this->wait_until(lock, timeout_time) == std::cv_status::timeout)
        return pred();
    return true;
  }
  template<class P>
  bool wait_for(std::unique_lock<Mutex>& lock, double duration, P pred)
  {
    return this->wait_until(lock, SIMIX_get_clock() + duration, std::move(pred));
  }

  // Notify functions

  void notify_one();
  void notify_all();

  XBT_ATTRIB_DEPRECATED("Use notify_one() instead")
  void notify() { notify_one(); }

private:
  smx_cond_t cond_;

};
}} // namespace simgrid::s4u

#endif /* SIMGRID_S4U_COND_VARIABLE_HPP */
